/* start of file Algo358.cpp ---------------------------------------------- */

char
 cname[]   = "Algo358 - The singular value decomposition of a complex matrix",
 cversion[]= "C++ Version 1.2, last revised: 2007-04-16",
 cauthor1[]= "Copyright (c) 2006-2007 by author: Harry J. Smith,",
 cauthor2[]= "19628 Via Monte Dr., Saratoga, CA 95070.  All rights reserved.";

/* <summary>
	SUBROUTINE CSVD - The singular value decomposition of a complex
	m by n matrix by ACM Algorithm 358
   </summary>
*/

#include "stdafx.h"
#include "Algo358.h"

using namespace std;

void CSVD(complex<float> a[][maxA], int m, int n, int p, int nu, int nv,
		float s[], complex<float> u[][maxA], complex<float> v[][maxA])
/* Singular Value Decomposition, a = u * s * Conj(Tran(v)), a is destroyed by CSVD
   the diagonal matrix s is output as a vector, m must be >= n
   if smaller, a should be filled with zero rows
   this code is adapted from Collected Algorithms from ACM, Algorithm 358
   The transformation Conj(Tran(u)) is applied to the p vectors given in columns
   n, n+1, ..., n+p-1 of matrix a
   See: http://www.scs.fsu.edu/~burkardt/f77_src/toms358/toms358.f
   and: http://www.scs.fsu.edu/~burkardt/f77_src/toms358/toms358_prb.f
*/
{
	float b[maxA];
	float c[maxA];
	float cs, eps, eta, f, g, h;
	int i, j, k, k1, L, L1, nM1, np;
	complex<float> q;
	complex<float> r;
	complex<float> cZero = complex<float>(0.0F, 0.0F);
	complex<float> cOne = complex<float>(1.0F, 0.0F);
	float sn;
	float t[maxA];
	float tol, w, x, y, z;
	eta = 1.5E-07F;			/* eta = the relative machine precision */
	tol = 1.5E-31F;			/* tol = the smallest normalized positive number, divided by eta */
	np = n + p;
	nM1 = n - 1;
	L = 0;
	/*
		HOUSEHOLDER REDUCTION
	*/
	c[0] = 0.0F;
	k = 0;
	while (1)
	{
		k1 = k + 1;
		/*
			ELIMINATION OF a[i][k], i = k, ..., m-1
		*/
		z = 0.0F;
		for (i = k; i < m; i++)
			z += norm(a[i][k]);
		b[k] = 0.0F;
		if (z > tol)
		{
			z = sqrt(z);
			b[k] = z;
			w = abs(a[k][k]);
			q = cOne;
			if (w != 0.0F) q = a[k][k] / w;
			a[k][k] = q * (z + w);
			if (k != np - 1)
			{
				for (j = k1; j < np; j++)
				{
					q = cZero;
					for (i = k; i < m; i++)
						q += conj(a[i][k]) * a[i][j];
					q /= z * (z + w);
					for (i = k; i < m; i++)
						a[i][j] -= q * a[i][k];
				}
			}
			/*
				PHASE TRANSFORMATION
			*/
			q = -conj(a[k][k]) / abs(a[k][k]);
			for (j = k1; j < np; j++)
				a[k][j] *= q;
		}
		/*
			ELIMINATION OF a[k][j], j = k+2, ..., n-1
		*/
		if (k == nM1) break;
		z = 0.0F;
		for (j = k1; j < n; j++)
			z += norm(a[k][j]);
		c[k1] = 0.0F;
		if (z > tol)
		{
			z = sqrt(z);
			c[k1] = z;
			w = abs(a[k][k1]);
			q = cOne;
			if (w != 0.0F) q = a[k][k1] / w;
			a[k][k1] = q * (z + w);
			for (i = k1; i < m; i++)
			{
				q = cZero;
				for (j = k1; j < n; j++)
					q += conj(a[k][j]) * a[i][j];
				q /= z * (z + w);
				for (j = k1; j < n; j++)
					a[i][j] -= q * a[k][j];
			}
			/*
				PHASE TRANSFORMATION
			*/
			q = -conj(a[k][k1]) / abs(a[k][k1]);
			for (i = k1; i < m; i++)
				a[i][k1] *= q;
		}
		k = k1;
	}
	/*
		TOLERANCE FOR NEGLIGIBLE ELEMENTS
	*/
	eps = 0.0F;
	for (k = 0; k < n; k++)
	{
		s[k] = b[k];
		t[k] = c[k];
		if (s[k] + t[k] > eps)
			eps = s[k] + t[k];
	}
	eps *= eta;
	/*
		INITIALIZATION OF u AND v
	*/
	if (nu > 0)
	{
		for (j = 0; j < nu; j++)
		{
			for (i = 0; i < m; i++)
				u[i][j] = cZero;
			u[j][j] = cOne;
		}
	}
	if (nv > 0)
	{
		for (j = 0; j < nv; j++)
		{
			for (i = 0; i < n; i++)
				v[i][j] = cZero;
			v[j][j] = cOne;
		}
	}
	/*
		QR DIAGONALIZATION
	*/
	for (k = nM1; k >= 0; k--)
	{
		/*
			TEST FOR SPLIT
		*/
		while (1)
		{
			for (L = k; L >= 0; L--)
			{
				if (abs(t[L]) <= eps) goto Test;
				if (abs(s[L - 1]) <= eps) break;
			}
			/*
				CANCELLATION OF E(L)
			*/
			cs = 0.0F;
			sn = 1.0F;
			L1 = L - 1;
			for (i = L; i <= k; i++)
			{
				f = sn * t[i];
				t[i] *= cs;
				if (abs(f) <= eps) goto Test;
				h = s[i];
				w = sqrt(f * f + h * h);
				s[i] = w;
				cs = h / w;
				sn = -f / w;
				if (nu > 0)
				{
					for (j = 0; j < n; j++)
					{
						x = real(u[j][L1]);
						y = real(u[j][i]);
						u[j][L1] = x * cs + y * sn;
						u[j][i] = y * cs - x * sn;
					}
				}
				if (np == n) continue;
				for (j = n; j < np; j++)
				{
					q = a[L1][j];
					r = a[i][j];
					a[L1][j] = q * cs + r * sn;
					a[i][j] = r * cs - q * sn;
				}
			}
			/*
				TEST FOR CONVERGENCE
			*/
	Test:	w = s[k];
			if (L == k) break;
			/*
				ORIGIN SHIFT
			*/
			x = s[L];
			y = s[k - 1];
			g = t[k - 1];
			h = t[k];
			f = ((y - w) * (y + w) + (g - h) * (g + h)) / (2.0F * h * y);
			g = sqrt(f * f + 1.0F);
			if (f < 0.0F) g = -g;
			f = ((x - w) * (x + w) + (y / (f + g) - h) * h) / x;
			/*
				QR STEP
			*/
			cs = 1.0F;
			sn = 1.0F;
			L1 = L + 1;
			for (i = L1; i <= k; i++)
			{
				g = t[i];
				y = s[i];
				h = sn * g;
				g = cs * g;
				w = sqrt(h * h + f * f);
				t[i - 1] = w;
				cs = f / w;
				sn = h / w;
				f = x * cs + g * sn;
				g = g * cs - x * sn;
				h = y * sn;
				y = y * cs;
				if (nv > 0)
				{
					for (j = 0; j < n; j++)
					{
						x = real(v[j][i - 1]);
						w = real(v[j][i]);
						v[j][i - 1] = x * cs + w * sn;
						v[j][i] = w * cs - x * sn;
					}
				}
				w = sqrt(h * h + f * f);
				s[i - 1] = w;
				cs = f / w;
				sn = h / w;
				f = cs * g + sn * y;
				x = cs * y - sn * g;
				if (nu > 0)
				{
					for (j = 0; j < n; j++)
					{
						y = real(u[j][i - 1]);
						w = real(u[j][i]);
						u[j][i - 1] = y * cs + w * sn;
						u[j][i] = w * cs - y * sn;
					}
				}
				if (n == np) continue;
				for (j = n; j < np; j++)
				{
					q = a[i - 1][j];
					r = a[i][j];
					a[i - 1][j] = q * cs + r * sn;
					a[i][j] = r * cs - q * sn;
				}
			}
			t[L] = 0.0F;
			t[k] = f;
			s[k] = x;
		}
		/*
			CONVERGENCE
		*/
		if (w >= 0.0F) continue;
		s[k] = -w;
		if (nv == 0) continue;
		for (j = 0; j < n; j++)
			v[j][k] = -v[j][k];
	}
	/*
		SORT SINGULAR VALUES
	*/
	for (k = 0; k < n; k++)	/* sort descending */
	{
		g = -1.0F;
		j = k;
		for (i = k; i < n; i++)	/* sort descending */
		{
			if (s[i] <= g) continue;
			g = s[i];
			j = i;
		}
		if (j == k) continue;
		s[j] = s[k];
		s[k] = g;
		if (nv > 0)
		{
			for (i = 0; i < n; i++)
			{
				q = v[i][j];
				v[i][j] = v[i][k];
				v[i][k] = q;
			}
		}
		if (nu > 0)
		{
			for (i = 0; i < n; i++)
			{
				q = u[i][j];
				u[i][j] = u[i][k];
				u[i][k] = q;
			}
		}
		if (n == np) continue;
		for (i = n; i < np; i++)
		{
			q = a[j][i];
			a[j][i] = a[k][i];
			a[k][i] = q;
		}
	}
	/*
		BACK TRANSFORMATION
	*/
	if (nu > 0)
	{
		for (k = nM1; k >= 0; k--)
		{
			if (b[k] == 0.0F) continue;
			q = -a[k][k] / abs(a[k][k]);
			for (j = 0; j < nu; j++)
				u[k][j] *= q;
			for (j = 0; j < nu; j++)
			{
				q = cZero;
				for (i = k; i < m; i++)
					q += conj(a[i][k]) * u[i][j];
				q /= abs(a[k][k]) * b[k];
				for (i = k; i < m; i++)
					u[i][j] -= q * a[i][k];
			}
		}
	}
	if (nv > 0 && n > 1)
	{
		for (k = n - 2; k >= 0; k--)
		{
			k1 = k + 1;
			if (c[k1] == 0.0F) continue;
			q = -conj(a[k][k1]) / abs(a[k][k1]);
			for (j = 0; j < nv; j++)
				v[k1][j] *= q;
			for (j = 0; j < nv; j++)
			{
				q = cZero;
				for (i = k1; i < n; i++)
					q += a[k][i] * v[i][j];
				q /= (abs(a[k][k1]) * c[k1]);
				for (i = k1; i < n; i++)
					v[i][j] -= q * conj(a[k][i]);
			}
		}
	}
} /* CSVD */

/* end of file Algo358.cpp ------------------------------------------------ */

