// EigenCF.cpp : Defines the entry point for the console application.

//				Tests the Eigen function

#include "stdafx.h"
#include "Algo358.h"

using namespace System;
using namespace System::Reflection;		// for Assembly. ...
using namespace std;

void Eigen(complex<float> a[][maxA], int n, complex<float> va[], complex<float> ve[][maxA], bool herm, bool *error);
void CSVD(complex<float> a[][maxA], int m, int n, int p, int nu, int nv,
		float s[], complex<float> u[][maxA], complex<float> v[][maxA]);
void TestEigen();
void TimeStamp();
void WriteMatrix(complex<float> [][maxA], int, int);
void WriteMatrixZ(complex<float> [][maxA], int, int);
void MatMulSq(int n, complex<float> a[][maxA], complex<float> b[][maxA], complex<float> p[][maxA]);
void Hermitian(complex<float> a[][maxA], int n);

int ti;		// test case number, 1 to 4
double time0;

// ------- initialize
void init(void)
{
	char name[]    = "EigenCF - Computes Eigenvalues and Eigenvectors of a square complex matrix";
	char author1[] = "Copyright (c) 1981-2007 by author: Harry J. Smith,";
	char author2[] = "19628 Via Monte Dr., Saratoga, CA 95070.  All rights reserved.";
	char usage1[]  = "Usage: EigenCF [1 to 4] [>Test.out]";
	String^ version;		// = "C++ Version x.x.xxxx.xxxxx"
	Version aVersion = Assembly::GetExecutingAssembly()->GetName()->Version->ToString();
	version = "C++ Console Version " + aVersion.ToString() + " ";
	printf("\n%s\n", name);		// show credits
	int len;
	len = version->LastIndexOf(" ");
	char sta[40];
	for (int i = 0; i < len; i++)
		sta[i] = (char)version[i];
	sta[len] = 0;
	printf("%s\n", sta);
	printf("%s\n", author1);
	printf("%s\n\n", author2);
	printf("%s\n\n", usage1);	// show usage
} // init

// The main entry point for the application EigenCF
//int _tmain(int argc, _TCHAR* argv[])
int main(int argc, char* argv[])
{
	init();
	char tc[5] = "1";
	if (argc > 1)							// if arguments
		strncpy_s(tc, 3, argv[1], 2);		// Copy string from user input line
	ti = atoi(tc);
	if (ti < 2 || ti > 4)
		ti = 1;
	printf("Running test case %i of 4\n\n", ti);
	time0 = Environment::TickCount / 1000.0;
	TimeStamp();
	printf("\nEigenCF:\n");
	TestEigen();
	printf("EigenCF:\n");
	printf("Normal end of execution\n\n");
	TimeStamp();
	return 0;
} // main

void TestEigen()		//test Eigen
{
	int i, j, n;
	bool herm;
	bool error;
	complex<float> a[maxA][maxA];
	complex<float> u[maxA][maxA];
	complex<float> v[maxA][maxA];
	complex<float> w[maxA][maxA];
	complex<float> va[maxA];
	complex<float> ve[maxA][maxA];
	complex<float> cZero = complex<float>(0.0F, 0.0F);
	a[0][0] = complex<float>( 0.4499F, -0.1267F);	// Set the matrix
	a[0][1] = complex<float>(-0.2361F,  0.0775F);
	a[0][2] = complex<float>( 0.5008F, -0.7799F);
	a[0][3] = complex<float>(-0.7702F, -0.3143F);
	a[0][4] = complex<float>(-0.5084F,  0.5621F);
	a[0][5] = complex<float>(-0.5577F, -0.0913F);
	a[0][6] = complex<float>(-0.7207F, -0.9798F);
	a[0][7] = complex<float>(-0.5367F,  0.4677F);
	a[1][0] = complex<float>(-0.8432F, -0.3443F);
	a[1][1] = complex<float>( 0.0186F, -0.6332F);
	a[1][2] = complex<float>( 0.3505F,  0.0166F);
	a[1][3] = complex<float>(-0.8892F,  0.2657F);
	a[1][4] = complex<float>(-0.5066F,  0.6005F);
	a[1][5] = complex<float>( 0.1421F,  0.0154F);
	a[1][6] = complex<float>(-0.2327F, -0.2606F);
	a[1][7] = complex<float>(-0.4535F,  0.4458F);
	a[2][0] = complex<float>( 0.5896F,  0.2601F);
	a[2][1] = complex<float>( 0.8928F,  0.0103F);
	a[2][2] = complex<float>( 0.4350F, -0.2666F);
	a[2][3] = complex<float>(-0.7799F, -0.5512F);
	a[2][4] = complex<float>(-0.1022F, -0.4500F);
	a[2][5] = complex<float>(-0.6975F, -0.8781F);
	a[2][6] = complex<float>( 0.3364F, -0.8415F);
	a[2][7] = complex<float>( 0.3890F,  0.6854F);
	a[3][0] = complex<float>( 0.3911F,  0.3234F);
	a[3][1] = complex<float>(-0.5605F,  0.7638F);
	a[3][2] = complex<float>(-0.2009F,  0.2707F);
	a[3][3] = complex<float>( 0.0314F, -0.4336F);
	a[3][4] = complex<float>(-0.1046F,  0.3267F);
	a[3][5] = complex<float>(-0.8820F, -0.9251F);
	a[3][6] = complex<float>( 0.4480F, -0.0054F);
	a[3][7] = complex<float>( 0.5823F, -0.2339F);
	a[4][0] = complex<float>(-0.1395F, -0.1561F);
	a[4][1] = complex<float>( 0.3064F,  0.0263F);
	a[4][2] = complex<float>(-0.0975F,  0.9019F);
	a[4][3] = complex<float>( 0.2491F,  0.5787F);
	a[4][4] = complex<float>( 0.4106F, -0.8102F);
	a[4][5] = complex<float>(-0.7131F,  0.5675F);
	a[4][6] = complex<float>( 0.0759F,  0.1507F);
	a[4][7] = complex<float>( 0.1801F, -0.3816F);
	a[5][0] = complex<float>(-0.5889F, -0.0238F);
	a[5][1] = complex<float>( 0.9861F, -0.4593F);
	a[5][2] = complex<float>(-0.5219F,  0.8539F);
	a[5][3] = complex<float>( 0.6382F, -0.0232F);
	a[5][4] = complex<float>( 0.8741F,  0.6109F);
	a[5][5] = complex<float>(-0.2208F, -0.1206F);
	a[5][6] = complex<float>(-0.0868F,  0.7716F);
	a[5][7] = complex<float>(-0.4976F,  0.4122F);
	a[6][0] = complex<float>(-0.3180F,  0.8621F);
	a[6][1] = complex<float>( 0.3905F, -0.8872F);
	a[6][2] = complex<float>( 0.8140F, -0.2542F);
	a[6][3] = complex<float>(-0.2290F,  0.4111F);
	a[6][4] = complex<float>(-0.9723F, -0.4355F);
	a[6][5] = complex<float>( 0.5195F,  0.4954F);
	a[6][6] = complex<float>(-0.8843F,  0.4552F);
	a[6][7] = complex<float>( 0.9156F, -0.5878F);
	a[7][0] = complex<float>( 0.1596F, -0.2962F);
	a[7][1] = complex<float>( 0.0272F,  0.3327F);
	a[7][2] = complex<float>(-0.5806F, -0.1952F);
	a[7][3] = complex<float>(-0.9280F, -0.5908F);
	a[7][4] = complex<float>( 0.6263F,  0.0384F);
	a[7][5] = complex<float>( 0.0210F, -0.7325F);
	a[7][6] = complex<float>(-0.9907F, -0.9717F);
	a[7][7] = complex<float>( 0.0878F, -0.9056F);
	if (ti == 1)
	{
		n = 5;
		herm = false;
	}
	if (ti == 2)
	{
		n = 5;
		herm = true;
		Hermitian(a, n);	// Set the matrix
	}
	if (ti == 3)
	{
		n = 8;
		herm = false;
	}
	if (ti == 4)
	{
		n = 8;
		Hermitian(a, n);	// Set the matrix
		herm = true;
	}
		printf("Call the Eigen function to compute the Eigenvalues and Eigenvectors of\n");
	if (herm)
		printf("an N by N Hermitian");
	else
		printf("an N by N complex");
	printf(" matrix A, N = %i.\n", n);
	printf("\nMatrix A:\n\n");
	WriteMatrix(a, n, n);
	int p = 0;
	int nu = n;
	int nv = n;

	// INITIAL TESTS
	void PolNorm(complex<float> a[], int *n);
	complex<float> pp[9];
	pp[0] = complex<float>( 0.0000F,  0.0000F);
	pp[1] = complex<float>(-0.2009F,  0.2707F);
	pp[2] = complex<float>( 0.4106F, -0.8102F);
	int ppn = 3;
	PolNorm(pp, &ppn);
	void Laguer(complex<float> a[], int n, int *its, complex<float> *x);
	pp[0] = complex<float>( 0.4499F, -0.1267F);
	pp[1] = complex<float>(-0.2009F,  0.2707F);
	pp[2] = complex<float>( 0.4106F, -0.8102F);
	int its;
	// answer 0.80826026, 1.0269444
	complex<float> root;
	bool polish = true;
	Laguer(pp, 3, &its, &root);
	void PolRoots(complex<float> p[], int n, bool polish, complex<float> r[]);
	complex<float> roots[2];
	PolRoots(pp, 3, polish, roots);
	void CharPoly(complex<float> a[][maxA], int n, complex<float> p[]);
	CharPoly(a, n, pp);
	// END OF INITIAL TESTS

	Eigen(a, n, va, ve, herm, &error);
	printf("Eigenvalues va:\n\n");
	for (i = 0; i < n; i++)
		printf("  (%.7f,%.7f)\n", real(va[i]), imag(va[i]));
	printf("\nEigenvectors ve:\n\n");
	WriteMatrix(ve, n, n);
	MatMulSq(n, a, ve, u);
	printf("A * ve:\n\n");
	WriteMatrix(u, n, n);
	for (i = 0; i < n; i++)
		for (j = 0; j < n; j++)
			w[i][j] = cZero;
	for (i = 0; i < n; i++)
		w[i][i] = va[i];
	MatMulSq(n, ve, w, v);
	complex<float> maxErr = cZero;
	for (i = 0; i < n; i++)
		for (j = 0; j < n; j++)
		{
			w[i][j] = u[i][j] - v[i][j];
			if (abs(real(w[i][j])) > real(maxErr))
				maxErr = complex<float>(abs(real(w[i][j])), imag(maxErr));
			if (abs(imag(w[i][j])) > imag(maxErr))
				maxErr = complex<float>(real(maxErr), abs(imag(w[i][j])));
		}
	printf("A * ve - ve * Diag(va), should be zero: Max error = (%.3e,%.3e)\n\n", real(maxErr), imag(maxErr));
	WriteMatrixZ(w, n, n);
} // TestEigen

void Hermitian(complex<float> a[][maxA], int n)
// make a an n by n Hermitian matrix. use upper triangle to make lower triangle
{
	for (int i = 0; i < n - 1; i++)
	{
		for (int j = i + 1; j < n; j++)
			a[j][i] = conj(a[i][j]);
	}
	for (int k = 0; k < n; k++)
		a[k][k] = complex<float>(real(a[k][k]), 0.0F);

} // Hermitian

void MatMulSq(int n, complex<float> a[][maxA], complex<float> b[][maxA], complex<float> p[][maxA])
// p = a * b. a, b, p all n by n complex matrices
{
	complex<float> cZero = complex<float>(0.0F, 0.0F);
	for (int i = 0; i < n; i++)
		for (int j = 0; j < n; j++)
		{
			p[i][j] = cZero;
			for (int k = 0; k < n; k++)
			{
				p[i][j] += a[i][k] * b[k][j];
			}
		}
} // MatMulSq

void WriteMatrix(complex<float> a[][maxA], int m, int n)
// print an m by n complex matrix
{
	for (int i = 0; i < m; i++)
	{
		for (int j = 0; j < n; j++)
		{
			printf(" ");
			if (real(a[i][j]) >= 0.0F)
				printf(" ");
			else
				printf("-");
			if (real(a[i][j]) == 0.0F)
				printf("0.0000");
			else
				printf("%.4f", abs(real(a[i][j])));
			if (imag(a[i][j]) >= 0.0F)
				printf("+");
			else
				printf("-");
			if (imag(a[i][j]) == 0.0F)
				printf("0.0000");
			else
				printf("%.4f", abs(imag(a[i][j])));
		}
		printf("\n");
	}
	printf("\n");
} // WriteMatrix

void WriteMatrixZ(complex<float> a[][maxA], int m, int n)
// print an m by n complex matrix that is near zero
{
	for (int i = 0; i < m; i++)
	{
		for (int j = 0; j < n; j++)
		{
			printf(" ");
			if (real(a[i][j]) >= 0.0F)
				printf(" ");
			else
				printf("-");
			if (real(a[i][j]) == 0.0F)
				printf("0.0000");
			else
				printf("%.0e", abs(real(a[i][j])));
			if (imag(a[i][j]) >= 0.0F)
				printf("+");
			else
				printf("-");
			if (imag(a[i][j]) == 0.0F)
				printf("0.0000");
			else
				printf("%.0e", abs(imag(a[i][j])));
		}
		printf("\n");
	}
	printf("\n");
} // WriteMatrixZ

void TimeStamp()
// printout the current MM/dd/yyyy hh:mm:ss AM/PM  s.ss sec. as a timestamp
{
	double time1 = Environment::TickCount / 1000.0; // TickCount = number of milliseconds since Windows started.
	double timeD = time1 - time0;
	time0 = time1;
	DateTime dt = DateTime::Now;
	String^ st = dt.ToString("MM/dd/yyyy hh:mm:ss tt");
	st += "  " + timeD.ToString("0.00") + " sec.";
	int len;
	len = 5 + st->LastIndexOf(" sec.");
	char sta[79];
	for (int i = 0; i < len; i++)
	{
		sta[i] = (char)st[i];
	}
	sta[len] = 0;
	printf("%s  %.2f sec.\n", sta, timeD);
} // TimeStamp

/* end of file EigenCF.cpp ------------------------------------------ */
